/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Data Visualization module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:GPL$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 or (at your option) any later version
** approved by the KDE Free Qt Foundation. The licenses are as published by
** the Free Software Foundation and appearing in the file LICENSE.GPL3
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "graphmodifier.h"
#include <QtDataVisualization/qcategory3daxis.h>
#include <QtDataVisualization/qvalue3daxis.h>
#include <QtDataVisualization/qbardataproxy.h>
#include <QtDataVisualization/q3dscene.h>
#include <QtDataVisualization/q3dcamera.h>
#include <QtDataVisualization/qbar3dseries.h>
#include <QtDataVisualization/q3dtheme.h>
#include <QtCore/QTime>
#include <QtWidgets/QComboBox>
#include <QtCore/qmath.h>
#include "ui_graphmodifier.h"
#include <vector>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>


using namespace std;
using namespace QtDataVisualization;

const QString celsiusString = QString(QChar(0xB0)) + "C";
const QString power2String = QString(QChar(0xB2));
const QString power3String = QString(QChar(0xB3));

GraphModifier::GraphModifier(Q3DBars *bargraph)
    : m_graph(bargraph),
      m_xRotation(0.0f),
      m_yRotation(0.0f),
      m_fontSize(5),
      m_segments(6),
      m_subSegments(2),
      m_minval(-1.0f),
      m_maxval(50.0f),

      m_valueAxis(new QValue3DAxis),
      m_dataAxis(new QCategory3DAxis),
      m_timeAxis(new QCategory3DAxis),
      m_Series(new QBar3DSeries),


      m_barMesh(QAbstract3DSeries::MeshBevelBar),
      m_smooth(false)
{

    m_graph->setShadowQuality(QAbstract3DGraph::ShadowQualitySoftMedium);
    m_graph->activeTheme()->setBackgroundEnabled(false);
    m_graph->activeTheme()->setFont(QFont("Times New Roman", m_fontSize));
    m_graph->activeTheme()->setLabelBackgroundEnabled(false);
    m_graph->setMultiSeriesUniform(true);
    m_graph->setBarSpacing(QSizeF(0.2,0.2));
    m_timeAxis->setRange(0,30);


    ifstream inFile;
    inFile.open("amussel_flash_test_data.txt");
    cout << "File opened!" << endl;

    stringstream strStream;
    strStream << inFile.rdbuf(); //read from file
    string str = strStream.str(); //str holds the content of the file
   // cout << str << endl;

    cout << "\n" << endl;

    if (inFile.fail()){
        cerr << "Error opening File!" << endl;
        }

    inFile.close(); // closing file

    // replacing strings "PRS", "TMP", TRB" with ints
    ReplaceStringInPlace(str, "PRS", "0"); // 0-pressure
    ReplaceStringInPlace(str, "TMP", "1"); // 1-temperature
    ReplaceStringInPlace(str, "TRB", "2"); // 2-turbidity


    stringstream ss(str);

    //Adding data to vector
    int i;
    while (ss >> i)
    {
        vectData.push_back(i);
        if (ss.peek() == ',')
            ss.ignore();
    }

    for (i=0; i< vectData.size(); i++)
       // cout << vectData.at(i) << endl;

    for (int i = 2; i < vectData.size(); i=i+90){
        listTime.append((double)vectData.at(i)/1000);
    }


    int j=0;
    int k=0;
    for (int i = 1; i < vectData.size(); i=i+3){
        DataList[j][k] = ((double)vectData.at(i) /100);
        if(j==2) {
            DataList[j][k] /= 10;
        }
        if(j==2){
            k++;
        }
        j=(j+1)%3;
     }


    for(int i=0; i<vectData.size()/9; i++)
    {
    m_time += QString::number(listTime[i]);

    }

    m_data << "∙10" + power2String + " [hPa] - Pressure" << "[" +  celsiusString + "] - Temperature" << "∙10" + power3String + " - Turbidity";

    m_valueAxis->setTitle("Data Value");
    m_valueAxis->setSegmentCount(m_segments);
    m_valueAxis->setSubSegmentCount(m_subSegments);
    m_valueAxis->setRange(m_minval, m_maxval);
    m_valueAxis->setLabelFormat(QString(QStringLiteral("%.3f ")));
    m_valueAxis->setLabelAutoRotation(30.0f);
    m_valueAxis->setTitleVisible(true);

    m_dataAxis->setTitle("Data Type");
    m_dataAxis->setLabelAutoRotation(30.0f);
    m_dataAxis->setTitleVisible(true);
    m_timeAxis->setTitle("Time [s]");
    m_timeAxis->setLabelAutoRotation(30.0f);
    m_timeAxis->setTitleVisible(true);


    m_graph->setValueAxis(m_valueAxis);
    m_graph->setRowAxis(m_dataAxis);
    m_graph->setColumnAxis(m_timeAxis);



    m_Series->setItemLabelFormat(QStringLiteral("Amussel Data at @colLabel [s]:  @valueLabel @rowLabel"));
    m_Series->setMesh(QAbstract3DSeries::MeshBevelBar);
    m_Series->setMeshSmooth(false);


    m_graph->addSeries(m_Series);

    changePresetCamera();

    resetTemperatureData();


    Q3DCamera *camera = m_graph->scene()->activeCamera();
    m_defaultAngleX = camera->xRotation();
    m_defaultAngleY = camera->yRotation();
    m_defaultZoom = camera->zoomLevel();
    m_defaultTarget = camera->target();

    m_animationCameraX.setTargetObject(camera);
    m_animationCameraY.setTargetObject(camera);
    m_animationCameraZoom.setTargetObject(camera);
    m_animationCameraTarget.setTargetObject(camera);

    m_animationCameraX.setPropertyName("xRotation");
    m_animationCameraY.setPropertyName("yRotation");
    m_animationCameraZoom.setPropertyName("zoomLevel");
    m_animationCameraTarget.setPropertyName("target");

    int duration = 1700;
    m_animationCameraX.setDuration(duration);
    m_animationCameraY.setDuration(duration);
    m_animationCameraZoom.setDuration(duration);
    m_animationCameraTarget.setDuration(duration);

    // The zoom always first zooms out above the graph and then zooms in
    qreal zoomOutFraction = 0.3;
    m_animationCameraX.setKeyValueAt(zoomOutFraction, QVariant::fromValue(0.0f));
    m_animationCameraY.setKeyValueAt(zoomOutFraction, QVariant::fromValue(90.0f));
    m_animationCameraZoom.setKeyValueAt(zoomOutFraction, QVariant::fromValue(50.0f));
    m_animationCameraTarget.setKeyValueAt(zoomOutFraction,
                                          QVariant::fromValue(QVector3D(0.0f, 0.0f, 0.0f)));

}


GraphModifier::~GraphModifier()
{
    delete m_graph;
}

void GraphModifier::resetTemperatureData()
{




    // Create data arrays
    QBarDataArray *dataSet = new QBarDataArray;
    QBarDataRow *dataRow;

    dataSet->reserve(m_data.size());
    for (int row = 0; row < m_data.size(); row++) {
        // Create a data row
        dataRow = new QBarDataRow(m_time.size());
        for (int column = 0, l = 0; column < m_time.size(); column=column+10, l++) {
            (*dataRow)[l].setValue(DataList[row][column]);
            }
        // Add the row to the set
        dataSet->append(dataRow);
    }

    // Add data to the data proxy (the data proxy assumes ownership of it)
    m_Series->dataProxy()->resetArray(dataSet, m_data, m_time);

}

void GraphModifier::ReplaceStringInPlace(string& subject, const string& search, const string& replace) {
    size_t pos = 0;
    while ((pos = subject.find(search, pos)) != string::npos) {
         subject.replace(pos, search.length(), replace);
         pos += replace.length();
    }
}

void GraphModifier::changeRange(int range)
{
    if (range >= m_data.count())
        m_dataAxis->setRange(0, m_data.count() - 1);
    else
        m_dataAxis->setRange(range, range);
}

void GraphModifier::changeStyle(int style)
{
    QComboBox *comboBox = qobject_cast<QComboBox *>(sender());
    if (comboBox) {
        m_barMesh = QAbstract3DSeries::Mesh(comboBox->itemData(style).toInt());
        m_Series->setMesh(m_barMesh);
    }
}

void GraphModifier::changePresetCamera()
{
    m_animationCameraX.stop();
    m_animationCameraY.stop();
    m_animationCameraZoom.stop();
    m_animationCameraTarget.stop();

    // Restore camera target in case animation has changed it
    m_graph->scene()->activeCamera()->setTarget(QVector3D(0.0f, 0.0f, 0.0f));


    static int preset = Q3DCamera::CameraPresetFront;

    m_graph->scene()->activeCamera()->setCameraPreset((Q3DCamera::CameraPreset)preset);

    if (++preset > Q3DCamera::CameraPresetDirectlyBelow)
        preset = Q3DCamera::CameraPresetFrontLow;

}

void GraphModifier::changeTheme(int theme)
{
    Q3DTheme *currentTheme = m_graph->activeTheme();
    currentTheme->setType(Q3DTheme::Theme(theme));
    emit backgroundEnabledChanged(currentTheme->isBackgroundEnabled());
    emit gridEnabledChanged(currentTheme->isGridEnabled());
    emit fontChanged(currentTheme->font());
    emit fontSizeChanged(currentTheme->font().pointSize());
}

void GraphModifier::changeLabelBackground()
{
    m_graph->activeTheme()->setLabelBackgroundEnabled(!m_graph->activeTheme()->isLabelBackgroundEnabled());
}

void GraphModifier::changeSelectionMode(int selectionMode)
{
    QComboBox *comboBox = qobject_cast<QComboBox *>(sender());
    if (comboBox) {
        int flags = comboBox->itemData(selectionMode).toInt();
        m_graph->setSelectionMode(QAbstract3DGraph::SelectionFlags(flags));
    }
}

void GraphModifier::changeFont(const QFont &font)
{
    QFont newFont = font;
    m_graph->activeTheme()->setFont(newFont);
}

void GraphModifier::changeFontSize(int fontsize)
{
    m_fontSize = fontsize;
    QFont font = m_graph->activeTheme()->font();
    font.setPointSize(m_fontSize);
    m_graph->activeTheme()->setFont(font);
}

void GraphModifier::shadowQualityUpdatedByVisual(QAbstract3DGraph::ShadowQuality sq)
{
    int quality = int(sq);
    // Updates the UI component to show correct shadow quality
    emit shadowQualityChanged(quality);
}

void GraphModifier::changeLabelRotation(int rotation)
{
    m_valueAxis->setLabelAutoRotation(float(rotation));
    m_timeAxis->setLabelAutoRotation(float(rotation));
    m_dataAxis->setLabelAutoRotation(float(rotation));
}

void GraphModifier::setAxisTitleVisibility(bool enabled)
{
    m_valueAxis->setTitleVisible(enabled);
    m_timeAxis->setTitleVisible(enabled);
    m_dataAxis->setTitleVisible(enabled);
}

void GraphModifier::setAxisTitleFixed(bool enabled)
{
    m_valueAxis->setTitleFixed(enabled);
    m_timeAxis->setTitleFixed(enabled);
    m_dataAxis->setTitleFixed(enabled);
}


void GraphModifier::zoomToSelectedBar()
{
    m_animationCameraX.stop();
    m_animationCameraY.stop();
    m_animationCameraZoom.stop();
    m_animationCameraTarget.stop();

    Q3DCamera *camera = m_graph->scene()->activeCamera();
    float currentX = camera->xRotation();
    float currentY = camera->yRotation();
    float currentZoom = camera->zoomLevel();
    QVector3D currentTarget = camera->target();

    m_animationCameraX.setStartValue(QVariant::fromValue(currentX));
    m_animationCameraY.setStartValue(QVariant::fromValue(currentY));
    m_animationCameraZoom.setStartValue(QVariant::fromValue(currentZoom));
    m_animationCameraTarget.setStartValue(QVariant::fromValue(currentTarget));

    QPoint selectedBar = m_graph->selectedSeries()
            ? m_graph->selectedSeries()->selectedBar()
            : QBar3DSeries::invalidSelectionPosition();

    if (selectedBar != QBar3DSeries::invalidSelectionPosition()) {
        // Normalize selected bar position within axis range to determine target coordinates
        QVector3D endTarget;
        float xMin = m_graph->columnAxis()->min();
        float xRange = m_graph->columnAxis()->max() - xMin;
        float zMin = m_graph->rowAxis()->min();
        float zRange = m_graph->rowAxis()->max() - zMin;
        endTarget.setX((selectedBar.y() - xMin) / xRange * 2.0f - 1.0f);
        endTarget.setZ((selectedBar.x() - zMin) / zRange * 2.0f - 1.0f);


        // Rotate the camera so that it always points approximately to the graph center

        qreal endAngleX = qAtan(qreal(endTarget.z() / endTarget.x())) / M_PI * -180.0 + 90.0;
        if (endTarget.x() > 0.0f)
            endAngleX -= 180.0f;
        float barValue = m_graph->selectedSeries()->dataProxy()->itemAt(selectedBar.x(),
                                                                        selectedBar.y())->value();
        float endAngleY = barValue >= 0.0f ? 30.0f : -30.0f;
        if (m_graph->valueAxis()->reversed())
            endAngleY *= -1.0f;


        m_animationCameraX.setEndValue(QVariant::fromValue(float(endAngleX)));
        m_animationCameraY.setEndValue(QVariant::fromValue(endAngleY));
        m_animationCameraZoom.setEndValue(QVariant::fromValue(250));  
        m_animationCameraTarget.setEndValue(QVariant::fromValue(endTarget));

    } else {
        // No selected bar, so return to the default view
        m_animationCameraX.setEndValue(QVariant::fromValue(m_defaultAngleX));
        m_animationCameraY.setEndValue(QVariant::fromValue(m_defaultAngleY));
        m_animationCameraZoom.setEndValue(QVariant::fromValue(m_defaultZoom));
        m_animationCameraTarget.setEndValue(QVariant::fromValue(m_defaultTarget));
    }

    m_animationCameraX.start();
    m_animationCameraY.start();
    m_animationCameraZoom.start();
    m_animationCameraTarget.start();
}


void GraphModifier::changeShadowQuality(int quality)
{
    QAbstract3DGraph::ShadowQuality sq = QAbstract3DGraph::ShadowQuality(quality);
    m_graph->setShadowQuality(sq);
    emit shadowQualityChanged(quality);
}


void GraphModifier::rotateX(int rotation)
{
    m_xRotation = rotation;
    m_graph->scene()->activeCamera()->setCameraPosition(m_xRotation, m_yRotation);
}

void GraphModifier::rotateY(int rotation)
{
    m_yRotation = rotation;
    m_graph->scene()->activeCamera()->setCameraPosition(m_xRotation, m_yRotation);
}


void GraphModifier::setBackgroundEnabled(int enabled)
{
    m_graph->activeTheme()->setBackgroundEnabled(bool(enabled));
}

void GraphModifier::setGridEnabled(int enabled)
{
    m_graph->activeTheme()->setGridEnabled(bool(enabled));
}

void GraphModifier::setSmoothBars(int smooth)
{
    m_smooth = bool(smooth);
    m_Series->setMeshSmooth(m_smooth);
}


void GraphModifier::setReverseValueAxis(int enabled)
{
    m_graph->valueAxis()->setReversed(enabled);
}

void GraphModifier::setReflection(bool enabled)
{
    m_graph->setReflection(enabled);
}
